from typing import Any, Dict
from typing_extensions import Self
from datetime import datetime
from zipfile import ZipFile

import os
import shutil
import json
import logging

from ...networking import NetworkManager, NetworkObject
from ...folder_management import FolderManager
from ...codable import KeyDescriptor


class Model(NetworkObject):

    """
        Represents a machine learning model object on Coretex.ai
        Contains properties that describe the model

        Properties
        ----------
        name : str
            model name
        createdById : str
            id of model
        createdOn : datetime
            date of model creation
        datasetId : int
            dataset id that is used for training the model
        spaceId : int
            space id that is used for training the model
        projectId : int
            project id that is used for training the model
        isTrained : bool
            True if model is trained, False otherwise
        isDeleted : bool
            True if model is deleted, False otherwise
        accuracy : float
            model accuracy
        experimentId : int
            experiment id of trained model
        meta : Dict[str, Any]
            model meta data
    """

    name: str
    createdById: str
    createdOn: datetime
    datasetId: int
    spaceId: int
    projectId: int
    isTrained: bool
    isDeleted: bool
    accuracy: float
    experimentId: int
    meta: Dict[str, Any]

    @classmethod
    def modelDescriptorFileName(cls) -> str:
        """
            Returns
            -------
            str -> name of model descriptor file
        """

        return "model_descriptor.json"

    @classmethod
    def _keyDescriptors(cls) -> Dict[str, KeyDescriptor]:
        descriptors = super()._keyDescriptors()
        descriptors["experimentId"] = KeyDescriptor("model_queue_id")

        return descriptors

    @classmethod
    def createModel(cls, name: str, experimentId: int, accuracy: float, meta: Dict[str, Any]) -> Self:
        """
            Creates Model object of the provided experiment with specified properties

            Parameters
            ----------
            name : str
                model name
            experimentId : int
                experiment id of model
            accuracy : float
                model accuracy
            meta : Dict[str, Any]
                model meta data

            Returns
            -------
            Self -> Model object

            Example
            -------
            >>> from coretex import Model, ExecutingExperiment
            \b
            >>> experiment = ExecutingExperiment.current()
            >>> model = Model.createModel(
                    name = experiment.name,
                    experimentId = experiment.id,
                    accuracy = 0.87,
                    meta = {}
                )
        """

        model = cls.create(parameters = {
            "name": name,
            "model_queue_id": experimentId,
            "accuracy": accuracy,
            "meta": meta
        })

        if model is None:
            raise ValueError(">> [Coretex] Failed to create Model entity")

        return model

    @classmethod
    def saveModelDescriptor(cls, path: str, contents: Dict[str, Any]) -> None:
        """
        Saves a model descriptor - a JSON file that provides a description of a
        machine learning model. It includes information such as the model's
        architecture, input and output shapes, labels, description and etc.

        Example
        -------
        >>> from coretex import ExecutingExperiment, Model
        >>> model = Model.createModel(experiment.name, experiment.id, accuracy, {})
        >>> model.saveModelDescriptor(modelPath, {
                "project_task": experiment.spaceTask,
                "labels": labels,
                "modelName": model.name,
                "description": experiment.description,

                "input_description":
                    Input shape is [x, y]

                    x is actually number of samples in dataset\n
                    y represents number of unique taxons for selected level in dataset,

                "input_shape": [x, y],

                "output_description":
                    Output shape - [x, z]

                    x is actually number of samples in dataset\n
                    z represents that output 2d array (table) is going to have only 1 column (1 prediction for each sample in dataset),

                "output_shape": [x, z]
            })
        """

        modelDescriptorPath = os.path.join(path, cls.modelDescriptorFileName())

        with open(modelDescriptorPath, "w", encoding = "utf-8") as file:
            json.dump(contents, file, ensure_ascii = False, indent = 4)

    def download(self) -> None:
        """
            Downloads and extracts the model zip file from Coretex.ai
        """

        if self.isDeleted or not self.isTrained:
            return

        modelZipDestination = os.path.join(FolderManager.instance().modelsFolder, f"{self.id}.zip")

        modelFolderDestination = os.path.join(FolderManager.instance().modelsFolder, f"{self.id}")
        if os.path.exists(modelFolderDestination):
            return

        os.mkdir(modelFolderDestination)

        response = NetworkManager.instance().genericDownload(
            endpoint=f"model/download?id={self.id}",
            destination=modelZipDestination
        )

        if response.hasFailed():
            logging.getLogger("coretexpylib").info(">> [Coretex] Failed to download the model")

        zipFile = ZipFile(modelZipDestination)
        zipFile.extractall(modelFolderDestination)
        zipFile.close()

    def upload(self, path: str) -> bool:
        """
            Uploads the model zip file to Coretex.ai

            Parameters
            ----------
            path : str
                Path to the saved model dir

            Returns
            -------
            bool -> True if model data uploaded successfully, False if model data upload has failed


            Example
            -------
            >>> from coretex import Model, ExecutingExperiment
            \b
            >>> experiment: ExecutingExperiment[NetworkDataset] = ExecutingExperiment.current()
            >>> model = Model.createModel(
                name = experiment.name,
                experimentId = experiment.id,
                accuracy = 0.87,
                meta = {}
            )
            >>> if not model.upload("path/to/model-dir"):
                    print("Failed to upload model")
        """

        if self.isDeleted:
            return False

        logging.getLogger("coretexpylib").info(">> [Coretex] Uploading model file...")

        shutil.make_archive(path, "zip", path)

        files = {
            ("file", open(f"{path}.zip", "rb"))
        }

        parameters = {
            "id": self.id
        }

        response = NetworkManager.instance().genericUpload("model/upload", files, parameters)
        if response.hasFailed():
            logging.getLogger("coretexpylib").info(">> [Coretex] Failed to upload model file")
        else:
            logging.getLogger("coretexpylib").info(">> [Coretex] Uploaded model file")

        return not response.hasFailed()
